var sys = require('sys');
var assert = require('assert');
var uglifyjs = require('uglify-js');
var typecheck = require('typecheck');

var JSPP = function(arg) {
    var _jsp = uglifyjs.parser;
    var _pro = uglifyjs.uglify;
    var _check = typecheck.getInstance().check;
    var _trues = (arg && arg.trues) || [];
    var _falses = (arg && arg.falses) || [];
    var _donts = (arg && arg.donts) || [];
    var _calc = (arg && arg.calc) || false;
    var _minify = (arg && arg.minify) || false;
    var _debug = (arg && arg.debug) || false;

    assert.ok(_check(_trues, "string*"));
    assert.ok(_check(_falses, "string*"));
    assert.ok(_check(_donts, "string*"));
    assert.ok(_check(_calc, "boolean"));
    assert.ok(_check(_minify, "boolean"));
    
    _trues.push("true");
    _falses.push("false", "null", "undefined", "NaN");

    var _inArray = function(v, a) {
        assert.ok(_check(a, "array"));
        for (var i = 0; i < a.length; i++)
            if (a[i] == v)
                return true;
        return false;
    };

    var _getName = function(n) {
        if (n[0] == "dot") {
            return _getName(n[1]) + "." + n[2];
        } else if (n[0] == "name") {
            return n[1];
        } else {
            return undefined;
        }
    };

    var _isTrue = function(n) {
        if (n[0] == "name" || n[0] == "dot") {
            var name = _getName(n);
            return _inArray(name, _trues);
        } else if (n[0] == "num" || n[0] == "string") {
            return _check(n[1], "true");
        } else {
            return false;
        }
    };

    var _isFalse = function(n) {
        if (n[0] == "name" || n[0] == "dot") {
            var name = _getName(n);
            return _inArray(name, _falses);
        } else if (n[0] == "num" || n[0] == "string") {
            return _check(n[1], "false");
        } else {
            return false;
        }
    };

    var _isNum = function(n) {
        return n[0] == "num";
    };

    var _isStr = function(n) {
        return n[0] == "string";
    };

    var that = {};

    that.preprocess = function(code) {
        if (!_check(code, "string"))
            throw new TypeError("invalid argument");

        var ast = _jsp.parse(code);
        var w = _pro.ast_walker();
        var astStr = sys.inspect(ast, null, null);
        var next = true;
        var count = 1;
        while (next) {
            if (_debug) {
                console.log("<" + count + ">");
                console.log(astStr + "\n");
                count++;
            }
            ast = w.with_walkers({
                "unary-prefix": function(op, expr) {
                    if (op == "!") {
                        if (_isTrue(expr))
                            return ["name", "false"];
                        if (_isFalse(expr))
                            return ["name", "true"];
                    }

                    return ["unary-prefix", op, w.walk(expr)];
                },
                "binary": function(op, left, right) {
                    if (op == "&&") {
                        if (_isFalse(left) || _isFalse(right))
                            return ["name", "false"];
                        if (_isTrue(left))
                            return w.walk(right);
                        if (_isTrue(right))
                            return w.walk(left);
                    }

                    if (op == "||") {
                        if (_isTrue(left) || _isTrue(right))
                            return ["name", "true"];
                        if (_isFalse(left))
                            return w.walk(right);
                        if (_isFalse(right))
                            return w.walk(left);
                    }

                    if (_calc && _isNum(left) && _isNum(right)) {
                        if (op == "+")
                            return ["num", left[1] + right[1]];
                        else if (op == "-")
                            return ["num", left[1] - right[1]];
                        else if (op == "*")
                            return ["num", left[1] * right[1]];
                        else if (op == "/")
                            return ["num", left[1] / right[1]];
                        else if (op == "%")
                            return ["num", left[1] % right[1]];
                        else if (op == "&")
                            return ["num", left[1] & right[1]];
                        else if (op == "|")
                            return ["num", left[1] | right[1]];
                        else if (op == "^")
                            return ["num", left[1] ^ right[1]];
                        else if (op == "<<")
                            return ["num", left[1] << right[1]];
                        else if (op == ">>")
                            return ["num", left[1] >> right[1]];
                        else if (op == ">>>")
                            return ["num", left[1] >>> right[1]];
                        else if (op == "<")
                            return ["name", left[1] < right[1] ? "true" : "false"];
                        else if (op == "<=")
                            return ["name", left[1] <= right[1] ? "true" : "false"];
                        else if (op == ">")
                            return ["name", left[1] > right[1] ? "true" : "false"];
                        else if (op == ">=")
                            return ["name", left[1] >= right[1] ? "true" : "false"];
                        else if (op == "==")
                            return ["name", left[1] == right[1] ? "true" : "false"];
                        else if (op == "!=")
                            return ["name", left[1] != right[1] ? "true" : "false"];
                    }

                    if (_calc && _isStr(left) && _isStr(right)) {
                        if (op == "+")
                            return ["string", left[1] + right[1]];
                        else if (op == "<")
                            return ["name", left[1] < right[1] ? "true" : "false"];
                        else if (op == "<=")
                            return ["name", left[1] <= right[1] ? "true" : "false"];
                        else if (op == ">")
                            return ["name", left[1] > right[1] ? "true" : "false"];
                        else if (op == ">=")
                            return ["name", left[1] >= right[1] ? "true" : "false"];
                        else if (op == "==")
                            return ["name", left[1] == right[1] ? "true" : "false"];
                        else if (op == "!=")
                            return ["name", left[1] != right[1] ? "true" : "false"];
                    }

                    return ["binary", op, w.walk(left), w.walk(right)];
                },
                "conditional": function(cond, t, e) {
                    if (_isTrue(cond))
                        return w.walk(t);
                    else if (_isFalse(cond))
                        return w.walk(e);

                    return [ "conditional", w.walk(cond), w.walk(t), w.walk(e) ];
                },
                "if": function(c, t, e) {
                    if (_isTrue(c)) {
                        return ["stat", w.walk(t)];
                    } else if (_isFalse(c)) {
                        if (e === undefined || e === null)
                            return ["block"];
                        else
                            return ["stat", w.walk(e)];
                    }

                    return ["if", w.walk(c), w.walk(t), w.walk(e)];
                },
                "call": function(e, as) {
                    if (e[0] == "name" && _inArray(e[1], _donts)) {
                        return ["block"];
                    } else {
                        return ["call", w.walk(e), as.map(w.walk)];
                    }
                }},
                function() {
                    return w.walk(ast);
                });

            var tmp = sys.inspect(ast, null, null);
            if (astStr == tmp)
                next = false;
            else
                astStr = tmp;
        }

        code = _pro.gen_code(ast);
        if (_minify) {
            ast = _jsp.parse(code);
            ast = _pro.ast_mangle(ast);
            ast = _pro.ast_squeeze(ast);
        }

        return _pro.gen_code(ast);
    };

    return that;
};

exports.create = JSPP;

